from tkinter import *
import math, time

NLEARN  = 10000 # 学習繰り返し回数
NSAMPLE = 6     # 訓練データサンプル数
NINPUT  = 3     # 入力ノード数
NHIDDEN = 5     # 中間層（隠れ層）の素子数
NOUTPUT = 1     # 出力素子数
EPS     = 0.3   # 学習時重み修正の程度
ENDERR  = 0.002 # プログラム停止のための学習誤差
def array1(N):
    return [0 for i in range(N)] 
def array2(N1, N2):
    return [[0 for i in range(N2)] for j in range(N1)]
#■トレーニング用データ
tx =[[0,0,0],[1,0,1],[1,1,1],[1,1,0],[1,0,0],[0,0,1]]
ty =[    [0],    [0],    [1],    [0],    [1],    [1]]
#■データ領域
x=array1(NINPUT+1); h = array1(NHIDDEN+1); y=array1(NOUTPUT)
w1        = array2(NINPUT+1, NHIDDEN+1); w2=array2(NHIDDEN+1, NOUTPUT+1) # 重み
h_back    = array1(NHIDDEN+1); y_back=array1(NOUTPUT+1)  # 隠れ素子, 出力素子の逆伝播量
sigmoidTB = array1(100)               # 計算高速化のためもシグモイド関数表
#■ 初期化
def initialize():
    for i in range(100): #シグモイド関数表作成
        sigmoidTB[i]= float(1.0/(1.0+math.exp(float((i-50)*(-0.2)))))
    for i in range(NINPUT+1):  #重み初期設定
        for j in range(NHIDDEN): w1[i][j]=0
    for i in range(i<NHIDDEN+1):
        for j in range(NOUTPUT): w2[i][j]=0
    w1[0][0] = 1; w2[0][0] =-1 #すべてを0にしない。
#■シグモイド関数表索引
def sigmoid(inDT):
    id= int(inDT*5+50)
    if   id > 99 : id = 99
    elif id <  0 : id =  0
    return sigmoidTB[id]
#■順伝播処理
def proceed(isample):
    for i in range(NINPUT): #入力値を第１段目に移す
        x[i]= float(tx[isample][i])
    x[NINPUT]=1.0
    for j in range(NHIDDEN) : #隠れ素子値の計算
        net=0.0
        for i in range(NINPUT+1): net += w1[i][j]*x[i]
        h[j]=sigmoid(net)
    for j in range(NOUTPUT): # 出力素子値の計算
        net=0.0
        for i in range(NHIDDEN+1): net += w2[i][j]*h[i]
        y[j]=sigmoid(net)
#■誤差評価
def evalError(isample):
    error=0
    for j in range(NOUTPUT):
        df = ty[isample][j]-y[j]
        error += df*df
    return error/ float(NOUTPUT)
#■逆伝播処理
def backProp(isample): 
    for j in range(NOUTPUT): # 出力差を、逆伝播初期値とする
        y_back[j]=(y[j]-ty[isample][j])*(1.0-y[j])*y[j]
        for i in range(NHIDDEN): # 隠れ層への逆伝播
            net=0.0
            for j in range(NOUTPUT): net += w2[i][j]*y_back[j]
            h_back[i]=net*(1.0-h[i])*h[i]
        for i in range(NINPUT+1): # 入力層の重み変更
            for j in range(NHIDDEN+1):
                w1[i][j] -= EPS*x[i]*h_back[j]
        for i in range(NHIDDEN+1): # 隠れ層の重み変更
            for j in range(NOUTPUT+1): w2[i][j] -= EPS*h[i]*y_back[j]
#■トレーニング結果表示
def dsp(canvas, i, error,N, merror, endErr):
    canvas.delete("DT")
    font=('Times',10); font2=('Times',7)
    canvas.create_text(200,10,text="学習回数 %5d  誤差評価値 %12.4e < %12.4e" % (i, merror, endErr),
                       fill="blue", font=font,tag="DT" )
    DX=math.log(10)*25; XX=200
    for i in range(4):#対数目盛
        canvas.create_text(XX,30,text="10" ,fill="black", font=font,tag="DT" )
        canvas.create_text(XX+10,25,text="%d" %(i-4),
                           fill="black", font=font2,tag="DT" )
        XX+=DX
        for j in range(1,10):
            x=200+math.log(math.pow(10,i)*float(j))*25
            canvas.create_line(x,40,x, 40+N*20, fill="blue", width=1,tag="DT" )

    x=200+math.log(endErr*10E3)*25
    canvas.create_line(x,40,x, 40+N*20, fill="darkred", width=1,tag="DT" )
    canvas.create_text(XX,30,text="10" ,fill="black", font=font,tag="DT" )
    canvas.create_text(XX+10,25,text="0", fill="black", font=font2,tag="DT" )
    
    for j in range(N):#訓練データのグラフ
        canvas.create_text(100,j*20+50, text="訓練データ No. %2d : (%12.4e) " %(j+1, error[j]),
                           fill="black", font=font,tag="DT")
        Y=50+j*20
        canvas.create_line(200,Y,200+math.log(error[j]*10E3)*25, Y, fill="red", width=5,tag="DT" )
    canvas.create_line(200,40,200, 40+N*20, fill="blue", width=1,tag="DT" )
    root.update()
    time.sleep(0.001)
#■メイン処理
root=Tk()
root.title("誤差逆伝播法(Back Propagation)")
canvas=Canvas(root, width=500, height=200)
canvas.pack()
root.update()
ERR=array1(NSAMPLE)
initialize()
for i in range(NLEARN):#学習回数分の繰返し
    merror=0
    for j in range(NSAMPLE):#各サンプルごとの学習
        proceed(j)                      # 前進処理
        ERR[j]=evalError(j)              # 誤差評価
        if ERR[j]>merror : merror=ERR[j]
        backProp(j)			# 逆伝播
    dsp(canvas,i,ERR,NSAMPLE, merror, ENDERR)
    if merror<ENDERR: break # 誤差最大値が終了誤差値より小さければ終了
root.mainloop()
